"""
This module provides the EvaluateModel class.
"""
import logging
import os
import warnings

import h5py
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import average_precision_score, roc_auc_score
from sklearn.metrics import mean_squared_error

from .sequences import Genome
from .utils import (PerformanceMetrics, _is_lua_trained_model,
                    initialize_logger, load_model_from_state_dict)
from .utils.methylation_performance_metrics import MethylationPerformanceMetrics

logger = logging.getLogger("selene")


class EvaluateMethylationModel(object):
    """
    Evaluate model on a test set of sequences with known targets.

    Parameters
    ----------
    model : torch.nn.Module
        The model architecture.
    criterion : torch.nn._Loss
        The loss function that was optimized during training.
    data_sampler : selene_sdk.samplers.Sampler
        Used to retrieve samples from the test set for evaluation.
    features : list(str)
        List of distinct features the model predicts.
    trained_model_path : str
        Path to the trained model file, saved using `torch.save`.
    output_dir : str
        The output directory in which to save model evaluation and logs.
    batch_size : int, optional
        Default is 64. Specify the batch size to process examples.
        Should be a power of 2.
    n_test_samples : int or None, optional
        Default is `None`. Use `n_test_samples` if you want to limit the
        number of samples on which you evaluate your model. If you are
        using a sampler of type `selene_sdk.samplers.OnlineSampler`,
        by default it will draw 640000 samples if `n_test_samples` is `None`.
    report_gt_feature_n_positives : int, optional
        Default is 10. In the final test set, each class/feature must have
        more than `report_gt_feature_n_positives` positive samples in order to
        be considered in the test performance computation. The output file that
        states each class' performance will report 'NA' for classes that do
        not have enough positive samples.
    use_cuda : bool, optional
        Default is `False`. Specify whether a CUDA-enabled GPU is available
        for torch to use during training.
    data_parallel : bool, optional
        Default is `False`. Specify whether multiple GPUs are available
        for torch to use during training.
    use_features_ord : list(str) or None, optional
        Default is None. Specify an ordered list of features for which to
        run the evaluation. The features in this list must be identical to or
        a subset of `features`, and in the order you want the resulting
        `test_targets.npz` and `test_predictions.npz` to be saved. If using
        a FileSampler or H5DataLoader for the evaluation, you can pass in
        a dataset with the targets matrix only containing these features, but
        note that this subsetted targets matrix MUST be ordered the same
        way as `features`, and the predictions and targets .npz output
        will be reordered according to `use_features_ord`.

    Attributes
    ----------
    model : torch.nn.Module
        The trained model.
    criterion : torch.nn._Loss
        The model was trained using this loss function.
    sampler : selene_sdk.samplers.Sampler
        The example generator.
    features : list(str)
        List of distinct features the model predicts.
    batch_size : int
        The batch size to process examples. Should be a power of 2.
    use_cuda : bool
        If `True`, use a CUDA-enabled GPU. If `False`, use the CPU.
    data_parallel : bool
        Whether to use multiple GPUs or not.
    metrics : dict
        A dictionary that maps metric names (`str`) to metric functions.
        By default, this contains `"roc_auc"`, which maps to
        `sklearn.metrics.roc_auc_score`, and `"average_precision"`,
        which maps to `sklearn.metrics.average_precision_score`.

    """

    def __init__(self,
                 model,
                 data_sampler,
                 features,
                 trained_model_path,
                 output_dir,
                 batch_size=64,
                 n_test_samples=None,
                 report_gt_feature_n_positives=10,
                 use_cuda=False,
                 data_parallel=False,
                 use_features_ord=None,
                 metrics=dict(roc_auc=roc_auc_score,
                              average_precision=average_precision_score)):
        self.cls_criterion = nn.BCELoss()
        self.reg_criterion = F.mse_loss

        trained_model = torch.load(
            trained_model_path, map_location=lambda storage, location: storage)
        if "state_dict" in trained_model:
            self.model = load_model_from_state_dict(
                trained_model["state_dict"], model)
        else:
            self.model = load_model_from_state_dict(
                trained_model, model)
        self.model.eval()

        self.sampler = data_sampler

        self.output_dir = output_dir
        os.makedirs(output_dir, exist_ok=True)

        self.features = np.array(features)
        self._use_ixs = list(range(len(features)))
        if use_features_ord is not None:
            feature_ixs = {f: ix for (ix, f) in enumerate(features)}
            self._use_ixs = []
            for f in use_features_ord:
                if f in feature_ixs:
                    self._use_ixs.append(feature_ixs[f])
                else:
                    warnings.warn(("Feature {0} in `use_features_ord` "
                                   "does not match any features in the list "
                                   "`features` and will be skipped.").format(f))
            self._write_features_ordered_to_file()

        initialize_logger(
            os.path.join(self.output_dir, "{0}.log".format(
                __name__)),
            verbosity=2)

        self.data_parallel = data_parallel
        if self.data_parallel:
            self.model = nn.DataParallel(model)
            logger.debug("Wrapped model in DataParallel")

        self.use_cuda = use_cuda
        if self.use_cuda:
            self.model.cuda()

        self.batch_size = batch_size

        metrics = dict(cls_average_precision=average_precision_score,)
        self._metrics = MethylationPerformanceMetrics(
            self._get_feature_from_index,
            report_gt_feature_n_positives=report_gt_feature_n_positives,
            metrics=metrics)

        self.n_test_samples = n_test_samples
        self._test_data, self._all_test_targets, self._all_test_inds = \
            self.sampler.get_data_and_targets(self.batch_size, n_test_samples)

        self._use_testmat_ixs = self._use_ixs[:]
        # if the targets shape is the same as the subsetted features,
        # reindex based on the subsetted list
        if self._all_test_targets.shape[1] == len(self._use_ixs):
            subset_features = {self.features[ix]: i for (i, ix) in
                               enumerate(sorted(self._use_ixs))}
            self._use_testmat_ixs = [
                subset_features[f] for f in self.features[self._use_ixs]]

        #self._all_test_targets = self._all_test_targets[
        #    :, self._use_testmat_ixs]

        # save the targets dataset now
        with h5py.File(os.path.join(self.output_dir, 'test_dataset.N={0}.h5'.format(
                self.n_test_samples)), 'a') as fh:
            if 'targets' in fh:
                del fh['targets']
                del fh['indicators']
            fh.create_dataset('targets', data=self._all_test_targets)
            fh.create_dataset('indicators', data=self._all_test_inds)

        # reset Genome base ordering when applicable.
        if (hasattr(self.sampler, "reference_sequence") and
                isinstance(self.sampler.reference_sequence, Genome)):
            if _is_lua_trained_model(model):
                Genome.update_bases_order(['A', 'G', 'C', 'T'])
            else:
                Genome.update_bases_order(['A', 'C', 'G', 'T'])

    def _write_features_ordered_to_file(self):
        """
        Write the feature ordering specified by `use_features_ord`
        after matching it with the `features` list from the class
        initialization parameters.
        """
        fp = os.path.join(self.output_dir, 'use_features_ord.txt')
        with open(fp, 'w+') as file_handle:
            for f in self.features: #[self._use_ixs]:
                file_handle.write('{0}\n'.format(f))

    def _get_feature_from_index(self, index):
        """
        Gets the feature at an index in the features list.

        Parameters
        ----------
        index : int

        Returns
        -------
        str
            The name of the feature/target at the specified index.

        """
        return self.features[index] #[self._use_ixs][index]

    def evaluate(self):
        """
        Passes all samples retrieved from the sampler to the model in
        batches and returns the predictions. Also reports the model's
        performance on these examples.

        Returns
        -------
        dict
            A dictionary, where keys are the features and the values are
            each a dict of the performance metrics (currently ROC AUC and
            AUPR) reported for each feature the model predicts.

        """
        batch_bce, batch_mse, batch_losses = [], [], []
        sequences = []
        all_predictions = []
        all_ind_predictions = []
        for (inputs, targets, inds) in self._test_data:
            sequences.append(inputs)
            inputs = torch.Tensor(inputs)
            targets = torch.Tensor(targets) #[:, self._use_testmat_ixs])
            inds = torch.Tensor(inds)
            if self.use_cuda:
                inputs = inputs.cuda()
                targets = targets.cuda()
                inds = inds.cuda()
            with torch.no_grad():
                predictions = None
                if _is_lua_trained_model(self.model):
                    predictions = self.model.forward(
                        inputs.transpose(1, 2).contiguous().unsqueeze_(2))
                else:
                    predictions = self.model.forward(
                        inputs.transpose(1, 2))
                predictions = predictions #[:, self._use_ixs]
                cls_pred = predictions[:, 0]
                reg_pred = predictions[:, 1:]
                cls_loss = self.cls_criterion(cls_pred, inds)

                reg_loss = 0
                if len(inds[inds == 1]) != 0:
                    subset_pred = reg_pred[inds == 1]
                    subset_tgt = targets[inds == 1]
                    mask = torch.isnan(subset_tgt)
                    reg_loss = self.reg_criterion(
                        subset_pred[~mask], subset_tgt[~mask])

                all_predictions.append(reg_pred)
                all_ind_predictions.append(cls_pred)
                batch_bce.append(cls_loss.item())
                total_loss = cls_loss
                if torch.is_tensor(reg_loss):
                    total_loss += reg_loss
                    batch_mse.append(reg_loss.item())
                batch_losses.append(total_loss.item())

        all_predictions = torch.vstack(all_predictions)
        all_ind_predictions = torch.hstack(all_ind_predictions)[:, None]
        all_predictions = all_predictions.data.cpu().numpy()
        all_ind_predictions = all_ind_predictions.data.cpu().numpy()

        cls_scores = self._metrics.update(
            all_ind_predictions, self._all_test_inds[:, None])
            #.reshape((len(all_ind_predictions), 1)))
        for name, score in cls_scores.items():
            logger.info("test {0}: {1}".format(name, score))

        np.savez_compressed(
            os.path.join(self.output_dir, "test_predictions.N={0}.npz".format(
                self.n_test_samples)),
            data=all_predictions)
        np.savez_compressed(
            os.path.join(self.output_dir, "test_predictions_inds.N={0}.npz".format(
                self.n_test_samples)),
            data=all_ind_predictions)

        with h5py.File(os.path.join(self.output_dir, 'test_dataset.N={0}.h5'.format(
                self.n_test_samples)), 'a') as fh:
            fh.create_dataset('sequences', data=np.vstack(sequences))

        lossdict = {
            'bce': np.average(batch_bce),
            'mse': np.average(batch_mse),
            'loss': np.average(batch_losses)
        }
        for name, score in lossdict.items():
            logger.info("test {0}: {1}".format(name, score))

        test_performance = os.path.join(
            self.output_dir, "test_performance.txt")
        feature_scores_dict = self._metrics.write_feature_scores_to_file(
            test_performance)

        return feature_scores_dict



"""
This module provides the MethylationDataSampler class.

TODO: Currently, only works with sequences from `selene_sdk.sequences.Genome`.
We would like to generalize this to `selene_sdk.sequences.Sequence` if possible.
"""
from collections import namedtuple
import logging
import random

import numpy as np
from numpy.random import default_rng

from functools import wraps
from .online_sampler import OnlineSampler
from ..utils import get_indices_and_probabilities

logger = logging.getLogger(__name__)


SampleIndices = namedtuple(
    "SampleIndices", ["indices", "weights"])
"""
A tuple containing the indices for some samples, and a weight to
allot to each index when randomly drawing from them.

TODO: this is common to both the intervals sampler and the
random positions sampler. Can we move this to utils or
somewhere else?

Parameters
----------
indices : list(int)
    The numeric index of each sample.
weights : list(float)
    The amount of weight assigned to each sample.

Attributes
----------
indices : list(int)
    The numeric index of each sample.
weights : list(float)
    The amount of weight assigned to each sample.

"""


class MethylationDataSampler(OnlineSampler):
    """This sampler randomly selects a position in the genome and queries for
    a sequence centered at that position for input to the model.

    TODO: generalize to selene_sdk.sequences.Sequence?

    Parameters
    ----------
    reference_sequence : selene_sdk.sequences.Genome
        A reference sequence from which to create examples.
    target_path : str
        Path to tabix-indexed, compressed BED file (`*.bed.gz`) of genomic
        coordinates mapped to the genomic features we want to predict.
    features : list(str)
        List of distinct features that we aim to predict.
    seed : int, optional
        Default is 436. Sets the random seed for sampling.
    validation_holdout : list(str) or float, optional
        Default is `['chr6', 'chr7']`. Holdout can be regional or
        proportional. If regional, expects a list (e.g. `['chrX', 'chrY']`).
        Regions must match those specified in the first column of the
        tabix-indexed BED file. If proportional, specify a percentage
        between (0.0, 1.0). Typically 0.10 or 0.20.
    test_holdout : list(str) or float, optional
        Default is `['chr8', 'chr9']`. See documentation for
        `validation_holdout` for additional information.
    sequence_length : int, optional
        Default is 1000. Model is trained on sequences of `sequence_length`
        where genomic features are annotated to the center regions of
        these sequences.
    center_bin_to_predict : int, optional
        Default is 200. Query the tabix-indexed file for a region of
        length `center_bin_to_predict`.
    feature_thresholds : float [0.0, 1.0], optional
        Default is 0.5. The `feature_threshold` to pass to the
        `GenomicFeatures` object.
    mode : {'train', 'validate', 'test'}
        Default is `'train'`. The mode to run the sampler in.
    save_datasets : list(str), optional
        Default is `['test']`. The list of modes for which we should
        save the sampled data to file.
    output_dir : str or None, optional
        Default is None. The path to the directory where we should
        save sampled examples for a mode. If `save_datasets` is
        a non-empty list, `output_dir` must be specified. If
        the path in `output_dir` does not exist it will be created
        automatically.

    Attributes
    ----------
    reference_sequence : selene_sdk.sequences.Genome
        The reference sequence that examples are created from.
    target : selene_sdk.targets.Target
        The `selene_sdk.targets.Target` object holding the features that we
        would like to predict.
    validation_holdout : list(str) or float
        The samples to hold out for validating model performance. These
        can be "regional" or "proportional". If regional, this is a list
        of region names (e.g. `['chrX', 'chrY']`). These regions must
        match those specified in the first column of the tabix-indexed
        BED file. If proportional, this is the fraction of total samples
        that will be held out.
    test_holdout : list(str) or float
        The samples to hold out for testing model performance. See the
        documentation for `validation_holdout` for more details.
    sequence_length : int
        The length of the sequences to  train the model on.
    modes : list(str)
        The list of modes that the sampler can be run in.
    mode : str
        The current mode that the sampler is running in. Must be one of
        the modes listed in `modes`.

    """
    def __init__(self,
                 reference_sequence,
                 target_path,
                 features,
                 seed=436,
                 validation_holdout=['chr6', 'chr7'],
                 test_holdout=['chr8', 'chr9'],
                 sequence_length=1000,
                 center_bin_to_predict=200,
                 feature_thresholds=0.5,
                 mode="train",
                 cache_modes=None,
                 save_datasets=[],
                 output_dir=None):
        super(MethylationDataSampler, self).__init__(
            reference_sequence,
            target_path,
            features,
            seed=seed,
            validation_holdout=validation_holdout,
            test_holdout=test_holdout,
            sequence_length=sequence_length,
            center_bin_to_predict=center_bin_to_predict,
            feature_thresholds=feature_thresholds,
            mode=mode,
            save_datasets=save_datasets,
            output_dir=output_dir)
        self._sample_from_mode = {}
        self._randcache = {}
        for mode in self.modes:
            self._sample_from_mode[mode] = None
            self._randcache[mode] = {"cache_indices": None, "sample_next": 0}
        self.sample_from_intervals = []
        self.interval_lengths = []

        self.cache_modes = cache_modes

        self.worker_id = 0
        self.train_rng = None #default_rng()
        self.eval_rng = default_rng()

        self._initialized = False
        self._reset_train = False

    def init(func):
        # delay initialization to allow  multiprocessing
        @wraps(func)
        def dfunc(self, *args, **kwargs):
            if not self._initialized:
                logger.info("Initialization of sampler with mode {0}, worker {1}".format(
                    self.mode, self.worker_id))

                if self.mode == "train" and self.train_rng is None:
                    self.seed += self.worker_id
                    np.random.seed(self.seed)
                    random.seed(self.seed + 1)
                    self.train_rng = default_rng()
                #elif self.mode != "train" and self.eval_rng is None:
                #    self.seed += self.worker_id
                #    np.random.seed(self.seed)
                #    random.seed(self.seed + 1)
                #    print('Initialize for rng {0}, {1}'.format(self.mode, self.seed))
                #    self.eval_rng = default_rng()

                if self._holdout_type == "chromosome":
                    self._partition_genome_by_chromosome()
                else:
                     self._partition_genome_by_proportion()
                if self.cache_modes is not None:
                    for mode in self.cache_modes:
                        self._update_randcache(mode=mode)
                else:
                    for mode in self.modes:
                        self._update_randcache(mode=mode)
                self._initialized = True
            if self._reset_train and self.mode == "train":
                self._update_randcache(mode=self.mode)
                self._reset_train = False
            return func(self, *args, **kwargs)
        return dfunc


    def _partition_genome_by_proportion(self):
        for chrom, len_chrom in self.reference_sequence.get_chr_lens():
            if '_' in chrom:
                continue
            if 'X' in chrom or 'Y' in chrom:
                continue
            self.sample_from_intervals.append(
                (chrom,
                 self.sequence_length,
                 len_chrom - self.sequence_length))
            self.interval_lengths.append(len_chrom)
        n_intervals = len(self.sample_from_intervals)

        select_indices = list(range(n_intervals))
        np.random.shuffle(select_indices)
        n_indices_validate = int(n_intervals * self.validation_holdout)
        val_indices, val_weights = get_indices_and_probabilities(
            self.interval_lengths, select_indices[:n_indices_validate])
        self._sample_from_mode["validate"] = SampleIndices(
            val_indices, val_weights)

        if self.test_holdout:
            n_indices_test = int(n_intervals * self.test_holdout)
            test_indices_end = n_indices_test + n_indices_validate
            test_indices, test_weights = get_indices_and_probabilities(
                self.interval_lengths,
                select_indices[n_indices_validate:test_indices_end])
            self._sample_from_mode["test"] = SampleIndices(
                test_indices, test_weights)

            tr_indices, tr_weights = get_indices_and_probabilities(
                self.interval_lengths, select_indices[test_indices_end:])
            self._sample_from_mode["train"] = SampleIndices(
                tr_indices, tr_weights)
        else:
            tr_indices, tr_weights = get_indices_and_probabilities(
                self.interval_lengths, select_indices[n_indices_validate:])
            self._sample_from_mode["train"] = SampleIndices(
                tr_indices, tr_weights)

    def _partition_genome_by_chromosome(self):
        for mode in self.modes:
            self._sample_from_mode[mode] = SampleIndices([], [])
        index = 0
        for (chrom, len_chrom) in self.reference_sequence.get_chr_lens():
            if '_' in chrom:
                continue
            if 'X' in chrom or 'Y' in chrom:
                continue
            if chrom in self.validation_holdout:
                self._sample_from_mode["validate"].indices.append(
                    index)
            elif self.test_holdout and chrom in self.test_holdout:
                self._sample_from_mode["test"].indices.append(
                    index)
            else:
                self._sample_from_mode["train"].indices.append(
                    index)

            self.sample_from_intervals.append(
                (chrom,
                 self.sequence_length,
                 len_chrom - self.sequence_length))
            self.interval_lengths.append(len_chrom - 2 * self.sequence_length)

        for mode in self.modes:
            sample_indices = self._sample_from_mode[mode].indices
            indices, weights = get_indices_and_probabilities(
                self.interval_lengths, sample_indices)
            self._sample_from_mode[mode] = \
                self._sample_from_mode[mode]._replace(
                    indices=indices, weights=weights)

    def _retrieve(self, chrom, position):
        try:
            bin_start = position - self._start_radius
            bin_end = position + self._end_radius
            indicator = self.target.is_positive(
                chrom, bin_start, bin_end)
            retrieved_targets = self.target.get_feature_data(
                chrom, bin_start, bin_end)
            window_start = position - self._start_window_radius
            window_end = position + self._end_window_radius
            if window_end - window_start < self.sequence_length:
                print(bin_start, bin_end,
                      self._start_radius, self._end_radius,
                      self._start_window_radius, self._end_window_radius,)
                return None
            if window_end - window_start != self.sequence_length:
                print(bin_start, bin_end,
                      self._start_radius, self._end_radius,
                      self._start_window_radius, self._end_window_radius,)
                return None
            strand = self.STRAND_SIDES[random.randint(0, 1)]
            retrieved_seq = \
                self.reference_sequence.get_encoding_from_coords(
                    chrom, window_start, window_end, strand)
            if retrieved_seq.shape[0] == self.sequence_length + 1:
                print("Sequence for ({0}, {1}) is 1 base too long ({0}, {2}, {3}, {4})".format(
                    chrom, position, window_start, window_end, strand))
                logger.info("Sequence for ({0}, {1}) is 1 base too long ({0}, {2}, {3}, {4})".format(
                    chrom, position, window_start, window_end, strand))
            if retrieved_seq.shape[0] == 0:
                #logger.info("Full sequence centered at {0} position {1} "
                #            "could not be retrieved. Sampling again.".format(
                #                chrom, position))
                return None
            elif np.mean(retrieved_seq==0.25) > 0.30:
                #print("Over 30% of the bases in the sequence centered "
                #      "at {0} position {1} are ambiguous ('N'). "
                #      "Sampling again.".format(chrom, position))
                #logger.info("Over 30% of the bases in the sequence centered "
                #            "at {0} position {1} are ambiguous ('N'). "
                #            "Sampling again.".format(chrom, position))
                return None
            if self.mode in self._save_datasets:
                feature_indices = ';'.join(
                    [str(f) for f in np.nonzero(retrieved_targets)[0]])
                self._save_datasets[self.mode].append(
                    [chrom,
                     window_start,
                     window_end,
                     strand,
                     feature_indices])
                if len(self._save_datasets[self.mode]) > 200000:
                    self.save_dataset_to_file(self.mode)
            return (retrieved_seq, retrieved_targets)
        except ValueError as e:
            print('ValueError: sampled ({0}, {1}) and got {2}'.format(chrom, position, e))
            logger.info('ValueError: sampled ({0}, {1}) and got {2}'.format(chrom, position, e))
            return None

    def _update_randcache(self, mode=None):
        if not mode:
            mode = self.mode

        if mode == 'train':
            self._update_seed()
        self._randcache[mode]["cache_indices"] = np.random.choice(
            self._sample_from_mode[mode].indices,
            size=1000000,
            replace=True,
            p=self._sample_from_mode[mode].weights)
        self._randcache[mode]["sample_next"] = 0

        print("Update the cache for mode {0} (seed={2}): {1}".format(
            mode, self._randcache[mode]['cache_indices'][:5], self.seed))
        logger.info("Update the cache for mode {0} (seed={2}): {1}".format(
            mode, self._randcache[mode]['cache_indices'][:5], self.seed))

    def _update_seed(self):
        self.seed += 1 + self.worker_id
        np.random.seed(self.seed)
        random.seed(self.seed + 1)
        self.train_rng = default_rng()

    def set_worker_id(self, worker_id):
        self.worker_id = worker_id

    @init
    def sample(self, batch_size=1, mode=None):
        """
        Randomly draws a mini-batch of examples and their corresponding
        labels.

        Parameters
        ----------
        batch_size : int, optional
            Default is 1. The number of examples to include in the
            mini-batch.
        mode : str, optional
            Default is None. The operating mode that the object should run in.
            If None, will use the current mode `self.mode`.

        Returns
        -------
        sequences, targets : tuple(numpy.ndarray, numpy.ndarray)
            A tuple containing the numeric representation of the
            sequence examples and their corresponding labels. The
            shape of `sequences` will be
            :math:`B \\times L \\times N`, where :math:`B` is
            `batch_size`, :math:`L` is the sequence length, and
            :math:`N` is the size of the sequence type's alphabet.
            The shape of `targets` will be :math:`B \\times F`,
            where :math:`F` is the number of features.

        """
        mode = mode if mode else self.mode
        sequences = np.zeros((batch_size, self.sequence_length, 4))
        targets = np.zeros((batch_size, self.n_features))
        indicator = np.zeros(batch_size)
        n_samples_drawn = 0
        while n_samples_drawn < batch_size:
            sample_index = self._randcache[mode]["sample_next"]
            if sample_index == len(self._randcache[mode]["cache_indices"]):
                self._update_randcache()
                sample_index = 0

            rand_interval_index = \
                self._randcache[mode]["cache_indices"][sample_index]
            self._randcache[mode]["sample_next"] += 1

            chrom, cstart, cend = \
                self.sample_from_intervals[rand_interval_index]
            if mode == 'train':
                position = self.train_rng.integers(cstart, cend, size=1)[0]
            else:
                position = self.eval_rng.integers(cstart, cend, size=1)[0]

            retrieve_output = self._retrieve(chrom, position)
            if not retrieve_output:
                continue
            if len(retrieve_output[0]) != self.sequence_length:
                logger.info("Error sampling ({0}, {1})".format(chrom, position))
                continue
            seq, seq_targets, methyl = retrieve_output
            #print("{2}: Sampled ({0}, {1}) {3} {4}".format(chrom, position, mode, self.seed, self.worker_id))

            sequences[n_samples_drawn, :, :] = seq
            targets[n_samples_drawn, :] = seq_targets
            indicator[n_samples_drawn] = methyl
            n_samples_drawn += 1
        return sequences, targets
        #return chrom, position, sequences, targets

    @init
    def sample_index(self, index, batch_size=1, mode=None):
        """
        Randomly draws a mini-batch of examples and their corresponding
        labels.

        Parameters
        ----------
        batch_size : int, optional
            Default is 1. The number of examples to include in the
            mini-batch.
        mode : str, optional
            Default is None. The operating mode that the object should run in.
            If None, will use the current mode `self.mode`.

        Returns
        -------
        sequences, targets : tuple(numpy.ndarray, numpy.ndarray)
            A tuple containing the numeric representation of the
            sequence examples and their corresponding labels. The
            shape of `sequences` will be
            :math:`B \\times L \\times N`, where :math:`B` is
            `batch_size`, :math:`L` is the sequence length, and
            :math:`N` is the size of the sequence type's alphabet.
            The shape of `targets` will be :math:`B \\times F`,
            where :math:`F` is the number of features.

        """
        mode = mode if mode else self.mode
        sequences = np.zeros((batch_size, self.sequence_length, 4))
        targets = np.zeros((batch_size, self.n_features))
        n_samples_drawn = 0
        while n_samples_drawn < batch_size:
            rand_interval_index = \
                self._randcache[mode]["cache_indices"][index]
            chrom, cstart, cend = \
                self.sample_from_intervals[rand_interval_index]
            position = self.rng.integers(cstart, cend, size=1)[0]
            #position = np.random.randint(cstart, cend)
            retrieve_output = self._retrieve(chrom, position)
            if not retrieve_output or len(retrieve_output[0]) != self.sequence_length:
                continue
            seq, seq_targets = retrieve_output
            sequences[n_samples_drawn, :, :] = seq
            targets[n_samples_drawn, :] = seq_targets
            n_samples_drawn += 1
        return (sequences, targets)



"""
This module provides the `SamplerDataLoader` and  `SamplerDataset` classes,
which allow parallel sampling for any Sampler using
torch DataLoader mechanism.
"""
import random
import  sys

import h5py
import numpy as np
import torch

from functools import wraps
from torch.utils.data import Dataset, DataLoader


class _SamplerDataset(Dataset):
    """
    This class provides a Dataset interface that wraps around a Sampler.
    `_SamplerDataset` is used internally by `SamplerDataLoader`.

    Parameters
    ----------
    sampler : selene_sdk.samplers.Sampler
        The sampler from which to draw data.

    Attributes
    ----------
    sampler : selene_sdk.samplers.Sampler
        The sampler from which to draw data.
    """
    def __init__(self, sampler):
        super(_SamplerDataset, self).__init__()
        self.sampler = sampler

    def __getitem__(self, index):
        """
        Retrieve sample(s) from self.sampler. Only index length affects the
        number of samples. The index values are not used.

        Parameters
        ----------
        index : int or any object with __len__ method implemented
            The size of index is used to determine the number of the
            samples to return.

        Returns
        ----------
        sequences, targets : tuple(numpy.ndarray, numpy.ndarray)
            A tuple containing the numeric representation of the
            sequence examples and their corresponding labels. The
            shape of `sequences` will be
            :math:`I \\times L \\times N`, where :math:`I` is
            `index`, :math:`L` is the sequence length, and
            :math:`N` is the size of the sequence type's alphabet.
            The shape of `targets` will be :math:`I \\times T`,
            where :math:`T` is the number of targets predicted.
        """
        sequences, targets = self.sampler.sample(
            batch_size=1 if isinstance(index, int) else len(index))
        if sequences.shape[0] == 1:
            sequences = sequences[0,:]
            targets = targets[0,:]
        return sequences, targets

    def __len__(self):
        """
        Implementing __len__ is required by the DataLoader. So as a workaround,
        this returns `sys.maxsize` which is a large integer which should
        generally prevent the DataLoader from reaching its size limit.

        Another workaround that is implemented is catching the StopIteration
        error while calling `next` and reinitialize the DataLoader.
        """
        return sys.maxsize


class SamplerDataLoader(DataLoader):
    """
    A DataLoader that provides parallel sampling for any `Sampler` object.
    `SamplerDataLoader` can be used with `MultiSampler` by specifying
    the `SamplerDataLoader` object as `train_sampler`, `validate_sampler`
    or `test_sampler` when initiating a `MultiSampler`.

    Parameters
    ----------
    sampler : selene_sdk.samplers.Sampler
        The sampler from which to draw data.
    num_workers : int, optional
        Default to 1. Number of workers to use for DataLoader.
    batch_size : int, optional
        Default to 1. The number of samples the iterator returns in one step.
    seed : int, optional
        Default to 436. The seed for random number generators.

    Attributes
    ----------
    sampler : selene_sdk.samplers.Sampler
        The sampler from which to draw data. Specified by the `sampler` param.
    num_workers : int
        Number of workers to use for DataLoader.
    batch_size : int
        The number of samples the iterator returns in one step.

    """
    def __init__(self,
                 sampler,
                 num_workers=1,
                 batch_size=1,
                 seed=436):
        def worker_init_fn(worker_id):
            """
            This function is called to initialize each worker with different
            numpy seeds (torch seeds are set by DataLoader automatically).
            """
            np.random.seed(seed + worker_id)

        args = {
            "batch_size": batch_size,
            "num_workers": num_workers,
            "pin_memory": True,
            "worker_init_fn": worker_init_fn
        }

        super(SamplerDataLoader, self).__init__(_SamplerDataset(sampler), **args)
        self.seed = seed


def unpackbits_sequence(sequence, s_len):
    sequence = np.unpackbits(sequence.astype(np.uint8), axis=-2)
    nulls = np.sum(sequence, axis=-1) == sequence.shape[-1]
    sequence = sequence.astype(float)
    sequence[nulls, :] = 1.0 / sequence.shape[-1]
    if sequence.ndim == 3:
        sequence = sequence[:, :s_len, :]
    else:
        sequence = sequence[:s_len, :]
    return sequence


def unpackbits_targets(targets, t_len):
    targets = np.unpackbits(targets, axis=-1).astype(float)
    if targets.ndim == 2:
        targets = targets[:, :t_len]
    else:
        targets = targets[:self.t_len]
    return targets


class _H5Dataset(Dataset):
    """
    This class provides a Dataset that directly loads sequences and targets
    from a hdf5 file. `_H5Dataset` is intended to be used internally by
    `H5DataLoader`.

    Parameters
    ----------
    file_path : str
        The file path of the hdf5 file.
    in_memory : bool, optional
        Default is False. If True, load entire dataset into memory.
    unpackbits : bool, optional
        Default is False. If True, unpack binary-valued array from uint8
        sequence and targets array. See `numpy.packbits` for details.
    sequence_key : str, optional
        Default is "sequences". Specify the name of the hdf5 dataset that contains
        sequence data.
    targets_key : str, optional
        Default is "targets". Specify the name of the hdf5 dataset that contains
        target data.

    Attributes
    ----------
    file_path : str
        The file path of the hdf5 file.
    in_memory : bool
        If True, load entire dataset into memory.
    unpackbits : bool
        If True, unpack binary-valued array from uint8
        sequence and targets array. See `numpy.packbits` for details.
    """
    def __init__(self,
                 file_path,
                 in_memory=False,
                 unpackbits=False,  # implies unpackbits for both
                 unpackbits_seq=False,
                 unpackbits_tgt=False,
                 sequence_key="sequences",
                 targets_key="targets",
                 use_seq_len=None,
                 shift=None):
        super(_H5Dataset, self).__init__()
        self.file_path = file_path
        self.in_memory = in_memory

        self.unpackbits = unpackbits
        self.unpackbits_seq = unpackbits_seq
        self.unpackbits_tgt = unpackbits_tgt

        self.use_seq_len = use_seq_len
        self.shift = shift
        self._seq_start, self._seq_end = None, None

        self._initialized = False
        self._sequence_key = sequence_key
        self._targets_key = targets_key

    def init(func):
        # delay initialization to allow multiprocessing
        @wraps(func)
        def dfunc(self, *args, **kwargs):
            if not self._initialized:
                self.db = h5py.File(self.file_path, 'r')

                if self.unpackbits:
                    self.s_len = self.db['{0}_length'.format(self._sequence_key)][()]
                    self.t_len = self.db['{0}_length'.format(self._targets_key)][()]
                elif self.unpackbits_seq:
                    self.s_len = self.db['{0}_length'.format(self._sequence_key)][()]
                elif self.unpackbits_tgt:
                    self.t_len = self.db['{0}_length'.format(self._targets_key)][()]

                if self.in_memory:
                    self.sequences = np.asarray(self.db[self._sequence_key])
                    self.targets = np.asarray(self.db[self._targets_key])
                else:
                    self.sequences = self.db[self._sequence_key]
                    self.targets = self.db[self._targets_key]

                self._initialized = True
            return func(self, *args, **kwargs)
        return dfunc

    @init
    def __getitem__(self, index):
        if isinstance(index, int):
            index = index % self.sequences.shape[0]
        sequence = self.sequences[index]
        targets = self.targets[index]

        if self.unpackbits:
            sequence = unpackbits_sequence(sequence, self.s_len)
            targets = unpackbits_targets(targets, self.t_len)
        elif self.unpackbits_seq:
            sequence = unpackbits_sequence(sequence, self.s_len)
        elif self.unpackbits_tgt:
            targets = unpackbits_targets(targets, self.t_len)

        if self._seq_start is None:
            self._seq_start = 0
            self._seq_end = len(sequence)

            if self.use_seq_len is not None:
                mid = len(sequence) // 2
                self._seq_start = int(mid - np.ceil(self.use_seq_len / 2))
                self._seq_end = mid + self.use_seq_len // 2
                if self.shift is not None:
                    self._seq_start += self.shift
                    self._seq_end += self.shift
        sequence = sequence[self._seq_start:self._seq_end]

        s = sequence.astype(np.float32)
        return (torch.from_numpy(s), torch.from_numpy(targets))


    @init
    def __len__(self):
        return self.sequences.shape[0]


class H5DataLoader(DataLoader):
    """
    H5DataLoader provides optionally parallel sampling from a HDF5
    dataset that contains sequences and targets data. The name of the
    array of sequences and targets data are specified by `sequence_key`
    and `targets_key` respectively. The sequences array should be
    of shape:math:`B \\times L \\times N`, where :math:`B` is
    the sample size, :math:`L` is the sequence length, and :math:`N` is
    the size of the sequence type's alphabet. The shape of the targets array
     will be :math:`B \\times F`, where :math:`F` is the number of features.

    H5DataLoader also supports compressed binary data (using `numpy.packbits`)
    with the `unpackbits` option. To generate compressed binary data, the
    sequences and targets array have to both be binary-valued, and then
    packed in the :math:`L` (sequence length) and `F` (number of features)
    dimensions, respectively.
    For the sequences array, represent unknown bases ("N"s) by binary
    data with all-ones in the encoding - they will be transformed to
    the correct representations in selene_sdk.sequences.Genome when unpacked.
    In addition, to unpack correctly, the length of the packed dimensions,
    i.e. :math:`L` and :math:`F` must be provided in two integer scalars
    named `{sequence_key}_length` and `{targets_key}_length` in the HDF5 file
    if `unpackbits==True`.

    You can generate a HDF5 dataset file by sampling from an online sampler
    using the script that we provided `scripts/write_sampled_h5.py`. An
    example config file template is in `config_examples/sample_h5.yml`.

    `H5DataLoader` can be used with `MultiSampler` by passing
    `SamplerDataLoader` object as `train_sampler`, `validate_sampler` or
    `test_sampler` when initiating a `MultiSampler`.

    Parameters
    ----------
    file_path : str
        The file path of the hdf5 file.
    in_memory : bool, optional
        Default is False. If True, load entire dataset into memory.
    num_workers : int, optional
        Default is 1. If greater than 1, use multiple processes to parallelize data
        sampling.
    use_subset : int, (int, int), list(int), or None, optional
        Default is None. If a single integer value is provided, sample from only
        the first `use_subset` rows of the dataset. If a tuple of integers is
        provided, `(<start-index>, <end-index>)`, sample from the range of
        rows specified. If a list of integers is provided, restrict to sampling
        only the indices specified in the list. If None, use the entire dataset.
    batch_size : int, optional
        Default is 1. Specify the batch size of the DataLoader.
    shuffle : bool, optional
        Default is True. If False, load the data in the original order.
    unpackbits : bool, optional
        Default is False. If True, unpack binary-valued array from uint8
        sequence and targets array. See `numpy.packbits` for details.
    sequence_key : str, optional
        Default is "sequences". Specify the name of the hdf5 dataset that contains
        sequence data.
    targets_key : str, optional
        Default is "targets". Specify the name of the hdf5 dataset that contains
        target data.

    Attributes
    ----------
    dataset : `_H5Dataset`
        The `_H5Dataset` to load data from.

    """
    def __init__(self,
                 dataset,
                 num_workers=1,
                 use_subset=None,
                 batch_size=1,
                 seed=436,
                 sampler=None,
                 batch_sampler=None,
                 shuffle=True):
        g = torch.Generator()
        g.manual_seed(seed)

        def worker_init_fn(worker_id):
            worker_seed = torch.initial_seed() % 2**32
            print("Worker seed", worker_seed)
            np.random.seed(worker_seed)
            random.seed(worker_seed)
            torch.manual_seed(worker_seed)

        args = {
            "batch_size": batch_size,
            "pin_memory": True,
            "worker_init_fn": worker_init_fn,
            "sampler": sampler,
            "batch_sampler": batch_sampler,
            "generator": g,
        }

        if hasattr(dataset, 'in_memory'):
            args['num_workers'] = 0 if dataset.in_memory else num_workers
        else:
            args['num_workers'] = num_workers

        if use_subset is not None:
            from torch.utils.data.sampler import SubsetRandomSampler
            if isinstance(use_subset, int):
                use_subset = list(range(use_subset))
            elif isinstance(use_subset, tuple) and len(use_subset) == 2:
                use_subset = list(range(use_subset[0], use_subset[1]))
            args["sampler"] = SubsetRandomSampler(use_subset)
        else:
            args["shuffle"] = shuffle

        super(H5DataLoader, self).__init__(dataset, **args)




"""
This module provides the `FileSampler` base class, which defines the
interface for classes that draw samples from a particular file type.
"""
from abc import ABCMeta
from abc import abstractmethod


class FileSampler(metaclass=ABCMeta):
    """
    Classes that implement `FileSampler` can be initialized
    in any way, but must implement the methods `sample`,
    `get_data_and_targets`, and `get_data`.
    """

    """
    The types of modes that the `Sampler` object can run in.
    """

    def __init__(self):
        """
        Constructs a new `FileSampler` object.
        """

    @abstractmethod
    def sample(self, batch_size=1):
        """
        Fetches a mini-batch of the data from the sampler.

        Parameters
        ----------
        batch_size : int, optional
            Default is 1. The size of the batch to retrieve.

        """
        raise NotImplementedError()

    @abstractmethod
    def get_data_and_targets(self, batch_size, n_samples):
        """
        This method fetches a subset of the sequence data and
        corresponding targets from the sampler, divided into batches.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int
            The total number of samples to retrieve.

        """
        raise NotImplementedError()

    @abstractmethod
    def get_data(self, batch_size, n_samples):
        """
        This method fetches a subset of the data from the sampler,
        divided into batches.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int
            The total number of samples to retrieve.

        """
        raise NotImplementedError()



"""
This module provides the RandomPositionsSampler class.

TODO: Currently, only works with sequences from `selene_sdk.sequences.Genome`.
We would like to generalize this to `selene_sdk.sequences.Sequence` if possible.
"""
from collections import namedtuple
import logging
import random

import numpy as np

from functools import wraps
from .online_sampler import OnlineSampler
from ..utils import get_indices_and_probabilities

logger = logging.getLogger(__name__)


SampleIndices = namedtuple(
    "SampleIndices", ["indices", "weights"])
"""
A tuple containing the indices for some samples, and a weight to
allot to each index when randomly drawing from them.

TODO: this is common to both the intervals sampler and the
random positions sampler. Can we move this to utils or
somewhere else?

Parameters
----------
indices : list(int)
    The numeric index of each sample.
weights : list(float)
    The amount of weight assigned to each sample.

Attributes
----------
indices : list(int)
    The numeric index of each sample.
weights : list(float)
    The amount of weight assigned to each sample.

"""


class MethylationDataSampler(OnlineSampler):
    """This sampler randomly selects a position in the genome and queries for
    a sequence centered at that position for input to the model.

    TODO: generalize to selene_sdk.sequences.Sequence?

    Parameters
    ----------
    reference_sequence : selene_sdk.sequences.Genome
        A reference sequence from which to create examples.
    targets : GenomicFeaturesH5
    features : list(str)
        List of distinct features that we aim to predict.
    seed : int, optional
        Default is 436. Sets the random seed for sampling.
    validation_holdout : list(str) or float, optional
        Default is `['chr6', 'chr7']`. Holdout can be regional or
        proportional. If regional, expects a list (e.g. `['chrX', 'chrY']`).
        Regions must match those specified in the first column of the
        tabix-indexed BED file. If proportional, specify a percentage
        between (0.0, 1.0). Typically 0.10 or 0.20.
    test_holdout : list(str) or float, optional
        Default is `['chr8', 'chr9']`. See documentation for
        `validation_holdout` for additional information.
    sequence_length : int, optional
        Default is 1000. Model is trained on sequences of `sequence_length`
        where genomic features are annotated to the center regions of
        these sequences.
    center_bin_to_predict : int, optional
        Default is 200. Query the tabix-indexed file for a region of
        length `center_bin_to_predict`.
    mode : {'train', 'validate', 'test'}
        Default is `'train'`. The mode to run the sampler in.
    save_datasets : list(str), optional
        Default is `['test']`. The list of modes for which we should
        save the sampled data to file.
    output_dir : str or None, optional
        Default is None. The path to the directory where we should
        save sampled examples for a mode. If `save_datasets` is
        a non-empty list, `output_dir` must be specified. If
        the path in `output_dir` does not exist it will be created
        automatically.

    Attributes
    ----------
    reference_sequence : selene_sdk.sequences.Genome
        The reference sequence that examples are created from.
    target : selene_sdk.targets.Target
        The `selene_sdk.targets.Target` object holding the features that we
        would like to predict.
    validation_holdout : list(str) or float
        The samples to hold out for validating model performance. These
        can be "regional" or "proportional". If regional, this is a list
        of region names (e.g. `['chrX', 'chrY']`). These regions must
        match those specified in the first column of the tabix-indexed
        BED file. If proportional, this is the fraction of total samples
        that will be held out.
    test_holdout : list(str) or float
        The samples to hold out for testing model performance. See the
        documentation for `validation_holdout` for more details.
    sequence_length : int
        The length of the sequences to  train the model on.
    modes : list(str)
        The list of modes that the sampler can be run in.
    mode : str
        The current mode that the sampler is running in. Must be one of
        the modes listed in `modes`.

    """
    def __init__(self,
                 reference_sequence,
                 target_path,
                 features,
                 seed=436,
                 validation_holdout=['chr6', 'chr7'],
                 test_holdout=['chr8', 'chr9'],
                 sequence_length=1000,
                 center_bin_to_predict=200,
                 feature_thresholds=0.5,
                 mode="train",
                 save_datasets=[],
                 output_dir=None):
        super(MethylationDataSampler, self).__init__(
            reference_sequence,
            target_path,
            features,
            seed=seed,
            validation_holdout=validation_holdout,
            test_holdout=test_holdout,
            sequence_length=sequence_length,
            center_bin_to_predict=center_bin_to_predict,
            feature_thresholds=feature_thresholds,
            mode=mode,
            save_datasets=save_datasets,
            output_dir=output_dir)

        self._sample_from_mode = {}
        self._randcache = {}
        for mode in self.modes:
            self._sample_from_mode[mode] = None
            self._randcache[mode] = {"cache_indices": None, "sample_next": 0}

        self.sample_from_intervals = []
        self.interval_lengths = []
        self._initialized = False

    def init(func):
        # delay initialization to allow  multiprocessing
        @wraps(func)
        def dfunc(self, *args, **kwargs):
            if not self._initialized:
                if self._holdout_type == "chromosome":
                    self._partition_genome_by_chromosome()
                else:
                     self._partition_genome_by_proportion()

                for mode in self.modes:
                    self._update_randcache(mode=mode)
                self._initialized = True
            return func(self, *args, **kwargs)
        return dfunc


    def _partition_genome_by_proportion(self):
        for chrom, len_chrom in self.reference_sequence.get_chr_lens():
            if '_' in chrom:
                continue
            if 'X' in chrom or 'Y' in chrom:
                continue
            self.sample_from_intervals.append(
                (chrom,
                 self.sequence_length,
                 len_chrom - self.sequence_length))
            self.interval_lengths.append(len_chrom)
        n_intervals = len(self.sample_from_intervals)

        select_indices = list(range(n_intervals))
        np.random.shuffle(select_indices)
        n_indices_validate = int(n_intervals * self.validation_holdout)
        val_indices, val_weights = get_indices_and_probabilities(
            self.interval_lengths, select_indices[:n_indices_validate])
        self._sample_from_mode["validate"] = SampleIndices(
            val_indices, val_weights)

        if self.test_holdout:
            n_indices_test = int(n_intervals * self.test_holdout)
            test_indices_end = n_indices_test + n_indices_validate
            test_indices, test_weights = get_indices_and_probabilities(
                self.interval_lengths,
                select_indices[n_indices_validate:test_indices_end])
            self._sample_from_mode["test"] = SampleIndices(
                test_indices, test_weights)

            tr_indices, tr_weights = get_indices_and_probabilities(
                self.interval_lengths, select_indices[test_indices_end:])
            self._sample_from_mode["train"] = SampleIndices(
                tr_indices, tr_weights)
        else:
            tr_indices, tr_weights = get_indices_and_probabilities(
                self.interval_lengths, select_indices[n_indices_validate:])
            self._sample_from_mode["train"] = SampleIndices(
                tr_indices, tr_weights)

    def _partition_genome_by_chromosome(self):
        for mode in self.modes:
            self._sample_from_mode[mode] = SampleIndices([], [])
        index = 0
        for (chrom, len_chrom) in self.reference_sequence.get_chr_lens():
            if '_' in chrom:
                continue
            if 'X' in chrom or 'Y' in chrom or 'M' in chrom:
                continue
            if chrom in self.validation_holdout:
                self._sample_from_mode["validate"].indices.append(
                    index)
            elif self.test_holdout and chrom in self.test_holdout:
                self._sample_from_mode["test"].indices.append(
                    index)
            else:
                self._sample_from_mode["train"].indices.append(
                    index)

            self.sample_from_intervals.append(
                (chrom,
                 self.sequence_length,
                 len_chrom - self.sequence_length))
            self.interval_lengths.append(len_chrom - 2 * self.sequence_length)
            index += 1

        for mode in self.modes:
            sample_indices = self._sample_from_mode[mode].indices
            indices, weights = get_indices_and_probabilities(
                self.interval_lengths, sample_indices)
            self._sample_from_mode[mode] = \
                self._sample_from_mode[mode]._replace(
                    indices=indices, weights=weights)

    def _retrieve(self, chrom, position, strand=None):
        bin_start = position - self._start_radius
        bin_end = position + self._end_radius
        #indicator = self.target.is_positive(
        #    chrom, bin_start, bin_end)
        indicator, retrieved_targets = self.target.get_feature_data(
            chrom, bin_start, bin_end)
        window_start = position - self._start_window_radius
        window_end = position + self._end_window_radius
        if window_end - window_start < self.sequence_length:
            print(bin_start, bin_end,
                  self._start_radius, self._end_radius,
                  self._start_window_radius, self._end_window_radius,)
            return None
        if strand is None:
            strand = self.STRAND_SIDES[random.randint(0, 1)]
        retrieved_seq = \
            self.reference_sequence.get_encoding_from_coords(
                chrom, window_start, window_end, strand)

        if retrieved_seq.shape[0] == 0:
            logger.info("Full sequence centered at {0} position {1} "
                        "could not be retrieved. Sampling again.".format(
                            chrom, position))
            return None
        elif np.mean(retrieved_seq==0.25) > 0.30:
            logger.info("Over 30% of the bases in the sequence centered "
                        "at {0} position {1} are ambiguous ('N'). "
                        "Sampling again.".format(chrom, position))
            return None

        if self.mode in self._save_datasets:
            feature_indices = ';'.join(
                [str(f) for f in np.nonzero(retrieved_targets)[0]])
            self._save_datasets[self.mode].append(
                [chrom,
                 window_start,
                 window_end,
                 strand,
                 feature_indices])
            if len(self._save_datasets[self.mode]) > 200000:
                self.save_dataset_to_file(self.mode)
        coord = (chrom, position, strand)
        return (retrieved_seq, retrieved_targets, indicator, '{0}_{1}_{2}'.format(chrom, position, strand))

    def _update_randcache(self, mode=None):
        if not mode:
            mode = self.mode
        self._randcache[mode]["cache_indices"] = np.random.choice(
            self._sample_from_mode[mode].indices,
            size=200000,
            replace=True,
            p=self._sample_from_mode[mode].weights)
        self._randcache[mode]["sample_next"] = 0

    @init
    def sample(self, batch_size=1, mode=None):
        """
        Randomly draws a mini-batch of examples and their corresponding
        labels.

        Parameters
        ----------
        batch_size : int, optional
            Default is 1. The number of examples to include in the
            mini-batch.
        mode : str, optional
            Default is None. The operating mode that the object should run in.
            If None, will use the current mode `self.mode`.

        Returns
        -------
        sequences, targets : tuple(numpy.ndarray, numpy.ndarray)
            A tuple containing the numeric representation of the
            sequence examples and their corresponding labels. The
            shape of `sequences` will be
            :math:`B \\times L \\times N`, where :math:`B` is
            `batch_size`, :math:`L` is the sequence length, and
            :math:`N` is the size of the sequence type's alphabet.
            The shape of `targets` will be :math:`B \\times F`,
            where :math:`F` is the number of features.

        """
        mode = mode if mode else self.mode
        sequences = np.zeros((batch_size, self.sequence_length, 4))
        targets = np.zeros((batch_size, self.n_features))
        indicator = np.zeros(batch_size)
        coords = []
        n_samples_drawn = 0
        while n_samples_drawn < batch_size:
            sample_index = self._randcache[mode]["sample_next"]
            if sample_index == len(self._randcache[mode]["cache_indices"]):
                self._update_randcache()
                sample_index = 0

            rand_interval_index = \
                self._randcache[mode]["cache_indices"][sample_index]
            self._randcache[mode]["sample_next"] += 1

            chrom, cstart, cend = \
                self.sample_from_intervals[rand_interval_index]
            position = np.random.randint(cstart, cend)

            retrieve_output = self._retrieve(chrom, position)
            if not retrieve_output:
                continue
            seq, seq_targets, methyl, coord = retrieve_output
            sequences[n_samples_drawn, :, :] = seq
            targets[n_samples_drawn, :] = seq_targets
            indicator[n_samples_drawn] = methyl
            coords.append(coord)
            n_samples_drawn += 1
        return (sequences, targets, indicator, coords)


    def get_data_and_targets(self, batch_size, n_samples=None, mode=None):
        """
        This method fetches a subset of the data from the sampler,
        divided into batches. This method also allows the user to
        specify what operating mode to run the sampler in when fetching
        the data.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int or None, optional
            Default is None. The total number of samples to retrieve.
            If `n_samples` is None and the mode is `validate`, will
            set `n_samples` to 32000; if the mode is `test`, will set
            `n_samples` to 640000 if it is None. If the mode is `train`
            you must have specified a value for `n_samples`.
        mode : str, optional
            Default is None. The mode to run the sampler in when
            fetching the samples. See
            `selene_sdk.samplers.IntervalsSampler.modes` for more
            information. If None, will use the current mode `self.mode`.

        Returns
        -------
        sequences_and_targets, targets_matrix : \
        tuple(list(tuple(numpy.ndarray, numpy.ndarray)), numpy.ndarray)
            Tuple containing the list of sequence-target pairs, as well
            as a single matrix with all targets in the same order.
            Note that `sequences_and_targets`'s sequence elements are of
            the shape :math:`B \\times L \\times N` and its target
            elements are of the shape :math:`B \\times F`, where
            :math:`B` is `batch_size`, :math:`L` is the sequence length,
            :math:`N` is the size of the sequence type's alphabet, and
            :math:`F` is the number of features. Further,
            `target_matrix` is of the shape :math:`S \\times F`, where
            :math:`S =` `n_samples`.

        """
        if mode is not None:
            self.set_mode(mode)
        else:
            mode = self.mode
        sequences_and_targets = []
        if n_samples is None and mode == "validate":
            n_samples = 32000
        elif n_samples is None and mode == "test":
            n_samples = 640000

        n_batches = int(n_samples / batch_size)
        for _ in range(n_batches):
            inputs, targets, indicator = self.sample(batch_size)
            sequences_and_targets.append((inputs, targets, indicator))
        targets_mat = np.vstack([t for (s, t, i) in sequences_and_targets])
        ind_mat = np.hstack([i for (s, t, i) in sequences_and_targets])
        if mode in self._save_datasets:
            self.save_dataset_to_file(mode, close_filehandle=True)
        return sequences_and_targets, targets_mat, ind_mat



"""
This module provides classes and methods for sampling labeled data
examples from files.
"""
from .file_sampler import FileSampler
from .bed_file_sampler import BedFileSampler
from .mat_file_sampler import MatFileSampler

__all__ = ["FileSampler",
           "BedFileSampler",
           "MatFileSampler"]



import os

from Cython.Distutils import build_ext
import numpy as np
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup


with open(os.path.join(os.path.dirname(__file__), "README.md"),
          encoding='utf-8') as readme:
    long_description = readme.read()

genome_module = Extension(
    "selene_sdk.sequences._sequence",
    ["selene_sdk/sequences/_sequence.pyx"],
    include_dirs=[np.get_include()])

genomic_features_module = Extension(
    "selene_sdk.targets._genomic_features",
    ["selene_sdk/targets/_genomic_features.pyx"],
    include_dirs=[np.get_include()])

ext_modules = [genome_module, genomic_features_module]
cmdclass = {'build_ext': build_ext}

setup(name="selene-sdk",
      version="0.6.0",
      long_description=long_description,
      long_description_content_type='text/markdown',
      description=("framework for developing sequence-level "
                   "deep learning networks"),
      packages=find_packages(),
      url="https://github.com/FunctionLab/selene",
      package_data={
        "selene_sdk.interpret": [
            "data/gencode_v28_hg38/*",
            "data/gencode_v28_hg19/*"
        ],
        "selene_sdk.sequences": [
            "data/*"
        ]
      },
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
      ],
      ext_modules=ext_modules,
      cmdclass=cmdclass,
      python_requires='>=3.9',
      install_requires=[
        "cython>=0.29.24",
        'click',
        "h5py",
        "matplotlib>=2.2.3",
        "numpy>=1.21.4",
        "pandas",
        "plotly",
        "pyfaidx",
        "pytabix",
        "pyyaml>=5.1",
        "ruamel.yaml",
        "scikit-learn",
        "scipy",
        "seaborn",
        "statsmodels",
        "torch>=1.0.0, <=2.3.1",
    ],
    entry_points={
        'console_scripts': [
            'selene_sdk = selene_sdk.cli:main',
        ],
    },
)



"""
This module provides the `TrainModel` class and supporting methods.
"""
import logging
import math
import os
import shutil
from time import strftime
from time import time

import numpy as np
import torch
import torch.autograd.profiler as profiler
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau
from scipy.stats import pearsonr
from scipy.stats import spearmanr
from sklearn.metrics import average_precision_score
from sklearn.metrics import roc_auc_score

from .utils import initialize_logger
from .utils import load_model_from_state_dict
from .utils.methylation_performance_metrics import MethylationPerformanceMetrics

logger = logging.getLogger("selene")


def _metrics_logger(name, out_filepath):
    logger = logging.getLogger("{0}".format(name))
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(message)s")
    file_handle = logging.FileHandler(
        os.path.join(out_filepath, "{0}.txt".format(name)))
    file_handle.setFormatter(formatter)
    logger.addHandler(file_handle)
    return logger


class TrainMethylationBinary(object):
    """
    This class ties together the various objects and methods needed to
    train and validate a model.

    TrainModel saves a checkpoint model (overwriting it after
    `save_checkpoint_every_n_steps`) as well as a best-performing model
    (overwriting it after `report_stats_every_n_steps` if the latest
    validation performance is better than the previous best-performing
    model) to `output_dir`.

    TrainModel also outputs 2 files that can be used to monitor training
    as Selene runs: `selene_sdk.train_model.train.txt` (training loss) and
    `selene_sdk.train_model.validation.txt` (validation loss & average
    ROC AUC). The columns in these files can be used to quickly visualize
    training history (e.g. you can use `matplotlib`, `plt.plot(auc_list)`)
    and see, for example, whether the model is still improving, if there are
    signs of overfitting, etc.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train.
    data_sampler : selene_sdk.samplers.Sampler
        The example generator.
    optimizer_class : torch.optim.Optimizer
        The optimizer to minimize loss with.
    optimizer_kwargs : dict
        The dictionary of keyword arguments to pass to the optimizer's
        constructor.
    batch_size : int
        Specify the batch size to process examples. Should be a power of 2.
    max_steps : int
        The maximum number of mini-batches to iterate over.
    report_stats_every_n_steps : int
        The frequency with which to report summary statistics. You can
        set this value to be equivalent to a training epoch
        (`n_steps * batch_size`) being the total number of samples
        seen by the model so far. Selene evaluates the model on the validation
        dataset every `report_stats_every_n_steps` and, if the model obtains
        the best performance so far (based on the user-specified loss function),
        Selene saves the model state to a file called `best_model.pth.tar` in
        `output_dir`.
    output_dir : str
        The output directory to save model checkpoints and logs in.
    save_checkpoint_every_n_steps : int or None, optional
        Default is 1000. If None, set to the same value as
        `report_stats_every_n_steps`
    save_new_checkpoints_after_n_steps : int or None, optional
        Default is None. The number of steps after which Selene will
        continually save new checkpoint model weights files
        (`checkpoint-<TIMESTAMP>.pth.tar`) every
        `save_checkpoint_every_n_steps`. Before this point,
        the file `checkpoint.pth.tar` is overwritten every
        `save_checkpoint_every_n_steps` to limit the memory requirements.
    n_validation_samples : int or None, optional
        Default is `None`. Specify the number of validation samples in the
        validation set. If `n_validation_samples` is `None` and the data sampler
        used is the `selene_sdk.samplers.IntervalsSampler` or
        `selene_sdk.samplers.RandomSampler`, we will retrieve 32000
        validation samples. If `None` and using
        `selene_sdk.samplers.MultiSampler`, we will use all
        available validation samples from the appropriate data file.
    n_test_samples : int or None, optional
        Default is `None`. Specify the number of test samples in the test set.
        If `n_test_samples` is `None` and

            - the sampler you specified has no test partition, you should not
              specify `evaluate` as one of the operations in the `ops` list.
              That is, Selene will not automatically evaluate your trained
              model on a test dataset, because the sampler you are using does
              not have any test data.
            - the sampler you use is of type `selene_sdk.samplers.OnlineSampler`
              (and the test partition exists), we will retrieve 640000 test
              samples.
            - the sampler you use is of type
              `selene_sdk.samplers.MultiSampler` (and the test partition
              exists), we will use all the test samples available in the
              appropriate data file.

    cpu_n_threads : int, optional
        Default is 1. Sets the number of OpenMP threads used for parallelizing
        CPU operations.
    use_cuda : bool, optional
        Default is `False`. Specify whether a CUDA-enabled GPU is available
        for torch to use during training.
    data_parallel : bool, optional
        Default is `False`. Specify whether multiple GPUs are available
        for torch to use during training.
    logging_verbosity : {0, 1, 2}, optional
        Default is 2. Set the logging verbosity level.

            * 0 - Only warnings will be logged.
            * 1 - Information and warnings will be logged.
            * 2 - Debug messages, information, and warnings will all be\
                  logged.

    checkpoint_resume : str or None, optional
        Default is `None`. If `checkpoint_resume` is not None, it should be the
        path to a model file generated by `torch.save` that can now be read
        using `torch.load`.
    use_scheduler : bool, optional
        Default is `True`. If `True`, learning rate scheduler is used to
        reduce learning rate on plateau. PyTorch ReduceLROnPlateau scheduler
        with patience=16 and factor=0.8 is used. Different scheduler parameters
        can be specified with `scheduler_kwargs`.
    deterministic : bool, optional
        Default is `False`. If `True`, will set
        `torch.backends.cudnn.deterministic` to True and
        `torch.backends.cudnn.benchmark = False`. In Selene CLI,
        if `random_seed` is set in the configuration YAML, Selene automatically
        passes in `deterministic=True` to the TrainModel class.
    scheduler_kwargs : dict, optional
        Default is patience=16, verbose=True, and factor=0.8. Set the parameters
        for the PyTorch ReduceLROnPlateau scheduler.
    stopping_criteria : list or None, optional
        Default is `None`. If `stopping_criteria` is not None, it should be a
        list specifying how to use early stopping. The first value should be
        a str corresponding to one of `metrics`. The second value should be an
        int indicating the patience. If the specified metric does not improve
        in the given patience (usually corresponding to the number of epochs),
        training stops early.

    Attributes
    ----------
    model : torch.nn.Module
        The model to train.
    sampler : selene_sdk.samplers.Sampler
        The example generator.
    criterion : torch.nn._Loss
        The loss function to optimize.
    optimizer : torch.optim.Optimizer
        The optimizer to minimize loss with.
    batch_size : int
        The size of the mini-batch to use during training.
    max_steps : int
        The maximum number of mini-batches to iterate over.
    nth_step_report_stats : int
        The frequency with which to report summary statistics.
    nth_step_save_checkpoint : int
        The frequency with which to save a model checkpoint.
    use_cuda : bool
        If `True`, use a CUDA-enabled GPU. If `False`, use the CPU.
    data_parallel : bool
        Whether to use multiple GPUs or not.
    output_dir : str
        The directory to save model checkpoints and logs.

    """

    def __init__(self,
                 model,
                 data_sampler,
                 optimizer_class,
                 optimizer_kwargs,
                 batch_size,
                 max_steps,
                 report_stats_every_n_steps,
                 output_dir,
                 save_checkpoint_every_n_steps=1000,
                 save_new_checkpoints_after_n_steps=None,
                 report_gt_feature_n_positives=10,
                 n_validation_samples=None,
                 n_test_samples=None,
                 cpu_n_threads=1,
                 use_cuda=False,
                 data_parallel=False,
                 logging_verbosity=2,
                 checkpoint_resume=None,
                 metrics=dict(roc_auc=roc_auc_score,
                              average_precision=average_precision_score),
                 use_scheduler=True,
                 deterministic=False,
                 scheduler_kwargs=dict(patience=16,
                                       verbose=True,
                                       factor=0.8),
                 stopping_criteria=None):
        """
        Constructs a new `TrainModel` object.
        """
        self.model = model
        self.sampler = data_sampler

        # these can be made parameters if desired
        self.cls_criterion = nn.BCELoss()

        self.optimizer = optimizer_class(
            self.model.parameters(), **optimizer_kwargs)

        self.batch_size = batch_size
        self.max_steps = max_steps
        self.nth_step_report_stats = report_stats_every_n_steps
        self.nth_step_save_checkpoint = None

        if not save_checkpoint_every_n_steps:
            self.nth_step_save_checkpoint = report_stats_every_n_steps
        else:
            self.nth_step_save_checkpoint = save_checkpoint_every_n_steps

        self._save_new_checkpoints = save_new_checkpoints_after_n_steps

        os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir

        initialize_logger(
            os.path.join(self.output_dir, "{0}.log".format(__name__)),
            verbosity=logging_verbosity)

        if deterministic:
            logger.info("Setting deterministic = True for reproducibility.")
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False


        logger.info("Training parameters set: batch size {0}, "
                    "number of steps per 'epoch': {1}, "
                    "maximum number of steps: {2}".format(
                        self.batch_size,
                        self.nth_step_report_stats,
                        self.max_steps))

        torch.set_num_threads(cpu_n_threads)

        self.use_cuda = use_cuda
        self.data_parallel = data_parallel

        if self.data_parallel:
            self.model = nn.DataParallel(model)
            logger.debug("Wrapped model in DataParallel")

        if self.use_cuda:
            self.model.cuda()
            logger.debug("Set modules to use CUDA")

        self._report_gt_feature_n_positives = report_gt_feature_n_positives
        self._metrics = dict(average_precision=average_precision_score,
                             roc_auc=roc_auc_score)
        self._n_validation_samples = n_validation_samples
        self._n_test_samples = n_test_samples
        self._use_scheduler = use_scheduler

        self._init_train(scheduler_kwargs)
        self._init_validate()
        if "test" in self.sampler.modes:
            self._init_test()
        if checkpoint_resume is not None:
            self._load_checkpoint(checkpoint_resume)

        if type(stopping_criteria) is list and len(stopping_criteria) == 2:
            stopping_metric, stopping_patience = stopping_criteria
            self._early_stopping = True
            if stopping_metric in self._metrics:
                self._stopping_metric = stopping_metric
                self._stopping_patience = stopping_patience
                self._stopping_reached = False
            else:
                logger.warning("Did not recognize stopping metric. Not performing early stopping.")
                self._early_stopping = False
        else:
            self._early_stopping = False

    def _load_checkpoint(self, checkpoint_resume):
        checkpoint = torch.load(
            checkpoint_resume,
            map_location=lambda storage, location: storage)
        if "state_dict" not in checkpoint:
            raise ValueError(
                ("'state_dict' not found in file {0} "
                 "loaded with method `torch.load`. Selene does not support "
                 "continued training of models that were not originally "
                 "trained using Selene.").format(checkpoint_resume))

        self.model = load_model_from_state_dict(
            checkpoint["state_dict"], self.model)

        self._start_step = checkpoint["step"]
        if self._start_step >= self.max_steps:
            self.max_steps += self._start_step

        self._min_loss = checkpoint["min_loss"]
        self.optimizer.load_state_dict(
            checkpoint["optimizer"])
        if self.use_cuda:
            for state in self.optimizer.state.values():
                for k, v in state.items():
                    if isinstance(v, torch.Tensor):
                        state[k] = v.cuda()

        logger.info(
            ("Resuming from checkpoint: step {0}, min loss {1}").format(
                self._start_step, self._min_loss))

    def _init_train(self, scheduler_kwargs):
        self._start_step = 0
        self._train_logger = _metrics_logger(
                "{0}.train".format(__name__), self.output_dir)
        self._train_logger.info("loss")
        if self._use_scheduler:
            self.scheduler = ReduceLROnPlateau(
                self.optimizer,
                'min',
                **scheduler_kwargs)
        self._time_per_step = []
        self._train_loss = []

    def _init_validate(self):
        self._min_loss = float("inf") # TODO: Should this be set when it is used later? Would need to if we want to train model 2x in one run.
        self._create_validation_set(n_samples=self._n_validation_samples)
        self._validation_metrics = MethylationPerformanceMetrics(
            self.sampler.get_feature_from_index,
            report_gt_feature_n_positives=self._report_gt_feature_n_positives,
            metrics=self._metrics)
        self._validation_logger = _metrics_logger(
                "{0}.validation".format(__name__), self.output_dir)

        self._validation_logger.info("\t".join(["loss",] +
            sorted([x for x in self._validation_metrics.metrics.keys()])))

    def _init_test(self):
        self._test_data = None
        self._n_test_samples = self._n_test_samples
        self._test_metrics = MethylationPerformanceMetrics(
            self.sampler.get_feature_from_index,
            report_gt_feature_n_positives=self._report_gt_feature_n_positives,
            metrics=self._metrics)

    def _create_validation_set(self, n_samples=None):
        """
        Generates the set of validation examples.

        Parameters
        ----------
        n_samples : int or None, optional
            Default is `None`. The size of the validation set. If `None`,
            will use all validation examples in the sampler.

        """
        logger.info("Creating validation dataset.")
        t_i = time()
        self._validation_data, self._all_validation_targets, self._all_validation_inds = \
            self.sampler.get_validation_set(
                self.batch_size, n_samples=n_samples)
        self._all_validation_preds = np.hstack([
            self._all_validation_inds.reshape(len(self._all_validation_inds), 1),
            self._all_validation_targets])
        t_f = time()
        logger.info(("{0} s to load {1} validation examples ({2} validation "
                     "batches) to evaluate after each training step.").format(
                      t_f - t_i,
                      len(self._validation_data) * self.batch_size,
                      len(self._validation_data)))

    def create_test_set(self):
        """
        Loads the set of test samples.
        We do not create the test set in the `TrainModel` object until
        this method is called, so that we avoid having to load it into
        memory until the model has been trained and is ready to be
        evaluated.

        """
        logger.info("Creating test dataset.")
        t_i = time()
        self._test_data, self._all_test_targets, self._all_test_inds = \
            self.sampler.get_test_set(
                self.batch_size, n_samples=self._n_test_samples)
        t_f = time()
        logger.info(("{0} s to load {1} test examples ({2} test batches) "
                     "to evaluate after all training steps.").format(
                      t_f - t_i,
                      len(self._test_data) * self.batch_size,
                      len(self._test_data)))
        np.savez_compressed(
            os.path.join(self.output_dir, "test_targets.npz"),
            data=self._all_test_targets)

    def _get_batch(self):
        """
        Fetches a mini-batch of examples

        Returns
        -------
        tuple(numpy.ndarray, numpy.ndarray)
            A tuple containing the examples and targets.

        """
        t_i_sampling = time()
        batch_sequences, batch_targets, batch_inds = self.sampler.sample(
            batch_size=self.batch_size)
        t_f_sampling = time()
        logger.debug(
            ("[BATCH] Time to sample {0} examples: {1} s.").format(
                 self.batch_size,
                 t_f_sampling - t_i_sampling))
        return (batch_sequences, batch_targets, batch_inds)

    def _checkpoint(self):
        checkpoint_dict = {
            "step": self.step,
            "arch": self.model.__class__.__name__,
            "state_dict": self.model.state_dict(),
            "min_loss": self._min_loss,
            "optimizer": self.optimizer.state_dict()
        }
        if self._save_new_checkpoints is not None and \
                self._save_new_checkpoints >= self.step:
            checkpoint_filename = "checkpoint-{0}".format(
                strftime("%m%d%H%M%S"))
            self._save_checkpoint(
                checkpoint_dict, False, filename=checkpoint_filename)
            logger.debug("Saving checkpoint `{0}.pth.tar`".format(
                checkpoint_filename))
        else:
            self._save_checkpoint(
                checkpoint_dict, False)

    def train_and_validate(self):
        """
        Trains the model and measures validation performance.

        """
        for step in range(self._start_step, self.max_steps):
            self.step = step
            self.train()

            if step % self.nth_step_save_checkpoint == 0:
                self._checkpoint()
            if self.step and self.step % self.nth_step_report_stats == 0:
                self.validate()
            if self._early_stopping and self._stopping_reached:
                logger.debug("Patience ran out. Stopping early.")
                break

        self.sampler.save_dataset_to_file("train", close_filehandle=True)

    def train(self):
        """
        Trains the model on a batch of data.

        Returns
        -------
        float
            The training loss.

        """
        t_i = time()
        self.model.train()
        self.sampler.set_mode("train")

        inputs, targets, inds = self._get_batch()
        inputs = torch.Tensor(inputs)
        targets = torch.Tensor(targets)
        inds = torch.Tensor(inds)

        if self.use_cuda:
            inputs = inputs.cuda()
            targets = targets.cuda()
            inds = inds.cuda()

        #start = torch.cuda.Event(enable_timing=True)
        #end = torch.cuda.Event(enable_timing=True)

        #start.record()
        #cls_pred, reg_pred = self.model(inputs.transpose(1, 2))
        pred = self.model(inputs.transpose(1, 2))
        #end.record()
        #torch.cuda.synchronize()
        #print("model time: {0}".format(start.elapsed_time(end)))
        targets = torch.hstack([inds, targets])
        mask = torch.isnan(targets)
        total_loss = self.cls_criterion(
            pred[~mask], targets[~mask])

        """
        cls_pred = pred[:, 0]
        reg_pred = pred[:, 1:]
        cls_loss = self.cls_criterion(cls_pred, inds.squeeze())

        reg_loss = 0
        inds = inds.ravel()
        if len(inds[inds == 1]) != 0:
            subset_pred = reg_pred[inds == 1]
            subset_tgt = targets[inds == 1]
            mask = torch.isnan(subset_tgt)
            reg_loss = self.reg_criterion(
                subset_pred[~mask], subset_tgt[~mask])
            # TODO: write a custom loss fn based on this?
            #reg_loss = self.reg_criterion(
            #    subset_pred, subset_tgt, reduction='none').nanmean(dim=1).mean()
        """
        self.optimizer.zero_grad()
        #total_loss = cls_loss #* 0.01
        #if torch.is_tensor(reg_loss):
        #    total_loss += reg_loss
        total_loss.backward()
        self.optimizer.step()

        self._train_loss.append(total_loss.item())
        t_f = time()

        self._time_per_step.append(t_f - t_i)
        if self.step and self.step % self.nth_step_report_stats == 0:
            logger.info(("[STEP {0}] average number "
                         "of steps per second: {1:.1f}").format(
                self.step, 1. / np.average(self._time_per_step)))
            self._train_logger.info(np.average(self._train_loss))
            logger.info("training loss: {0}".format(np.average(self._train_loss)))
            self._time_per_step = []
            self._train_loss = []


    def _evaluate_on_data(self, data_in_batches):
        """
        Makes predictions for some labeled input data.

        Parameters
        ----------
        data_in_batches : list(tuple(numpy.ndarray, numpy.ndarray))
            A list of tuples of the data, where the first element is
            the example, and the second element is the label.

        Returns
        -------
        tuple(float, list(numpy.ndarray))
            Returns the average loss, and the list of all predictions.

        """
        self.model.eval()

        batch_losses = []
        all_predictions = []
        for (seqs, targets, inds) in data_in_batches:
            seqs = torch.Tensor(seqs)
            targets = torch.Tensor(targets)
            inds = torch.Tensor(inds)
            if self.use_cuda:
                seqs = seqs.cuda()
                targets = targets.cuda()
                inds = inds.cuda()

            with torch.no_grad():
                #cls_pred, reg_pred = self.model(seqs.transpose(1, 2))
                pred = self.model(seqs.transpose(1, 2))
                targets = torch.hstack([inds.reshape(len(inds), 1), targets])
                mask = torch.isnan(targets)
                total_loss = self.cls_criterion(
                    pred[~mask], targets[~mask])

                """
                cls_pred = pred[:, 0]
                reg_pred = pred[:, 1:]
                cls_loss = self.cls_criterion(cls_pred, inds)

                reg_loss = 0
                if len(inds[inds == 1]) != 0:
                    subset_pred = reg_pred[inds == 1]
                    subset_tgt = targets[inds == 1]
                    mask = torch.isnan(subset_tgt)
                    reg_loss = self.reg_criterion(
                        subset_pred[~mask], subset_tgt[~mask])
                    #subset_pred = reg_pred[inds == 1]
                    #subset_tgt = targets[inds == 1]
                    #reg_loss = self.reg_criterion(
                    #    subset_pred, subset_tgt, reduction='none').nanmean(dim=1).mean()
                """
                all_predictions.append(pred)
                #all_ind_predictions.append(cls_pred)
                #batch_bce.append(cls_loss.item())

                #total_loss = cls_loss #* 0.01
                #if torch.is_tensor(reg_loss):
                #    total_loss += reg_loss
                #    batch_mse.append(reg_loss.item())

                batch_losses.append(total_loss.item())
        all_predictions = torch.vstack(all_predictions)
        #all_ind_predictions = torch.hstack(all_ind_predictions)[:, None]
        all_predictions = all_predictions.data.cpu().numpy()
        #all_ind_predictions = all_ind_predictions.data.cpu().numpy()

        loss_breakdown = {
            #'bce': np.average(batch_bce),
            #'mse': np.average(batch_mse),
            'loss': np.average(batch_losses)
        }
        return loss_breakdown, all_predictions

    def validate(self):
        """
        Measures model validation performance.

        Returns
        -------
        dict
            A dictionary, where keys are the names of the loss metrics,
            and the values are the average value for that metric over
            the validation set.

        """
        lossdict, val_pred_tgts = self._evaluate_on_data(
            self._validation_data)
        cls_valid_scores = self._validation_metrics.update(
            val_pred_tgts, self._all_validation_preds,
            scores=['average_precision', 'roc_auc'])
        #reg_valid_scores = self._validation_metrics.update(
        #    val_pred_tgts[self._all_validation_inds == 1],
        #    self._all_validation_targets[self._all_validation_inds == 1],
        #    scores=['pearsonr', 'spearmanr'])
        for name, score in cls_valid_scores.items():
            logger.info("validation {0}: {1}".format(name, score))
            lossdict[name] = score
        #for name, score in reg_valid_scores.items():
        #    logger.info("validation {0}: {1}".format(name, score))
        #    lossdict[name] = score

        to_log = [lossdict['loss'],] #lossdict['mse'], lossdict['bce']]
        for k in sorted(self._validation_metrics.metrics.keys()):
            if k in lossdict and lossdict[k]:
                to_log.append(lossdict[k])
            else:
                to_log.append("NA")
        self._validation_logger.info('\t'.join([str(s) for s in to_log]))
        validation_loss = lossdict['loss']
        # scheduler update
        if self._use_scheduler:
            self.scheduler.step(
                math.ceil(validation_loss * 1000.0) / 1000.0)

        # save best_model
        if validation_loss < self._min_loss:
            self._min_loss = validation_loss
            self._save_checkpoint({
                "step": self.step,
                "arch": self.model.__class__.__name__,
                "state_dict": self.model.state_dict(),
                "min_loss": self._min_loss,
                "optimizer": self.optimizer.state_dict()}, True)
            logger.debug("Updating `best_model.pth.tar`")
        logger.info("validation loss: {0}".format(validation_loss))

        # check for early stopping
        if self._early_stopping:
            stopping_metric = self._validation_metrics.metrics[self._stopping_metric].data
            index = np.argmax(stopping_metric)
            if self._stopping_patience - (len(stopping_metric) - index - 1) <= 0:
                self._stopping_reached = True

    def evaluate(self):
        """
        Measures the model test performance.

        Returns
        -------
        dict
            A dictionary, where keys are the names of the loss metrics,
            and the values are the average value for that metric over
            the test set.

        """
        if self._test_data is None:
            self.create_test_set()
        losdict, test_pred_inds, test_pred_tgts = self._evaluate_on_data(
            self._test_data)

        test_scores = self._test_metrics.update(
            test_pred_inds, self._all_test_inds.reshape(len(test_pred_inds), 1),
            scores=['cls_average_precision'])

        #reg_test_scores = self._test_metrics.update(
        #    test_pred_tgts[self._all_test_inds == 1],
        #    self._all_test_targets[self._all_test_inds == 1],
        #    scores=['reg_mse'])

        np.savez_compressed(
            os.path.join(self.output_dir, "test_predictions.npz"),
            data=test_pred_tgts)
        np.savez_compressed(
            os.path.join(self.output_dir, "test_predictions.indicator.npz"),
            data=test_pred_inds)

        for name, score in test_scores.items():
            logger.info("test {0}: {1}".format(name, score))
            lossdict[name] = score
        #for name, score in reg_test_scores.items():
        #    logger.info("test {0}: {1}".format(name, score))
        #    lossdict[name] = score

        test_performance = os.path.join(
            self.output_dir, "test_performance.txt")
        feature_scores_dict = self._test_metrics.write_feature_scores_to_file(
            test_performance)

        return (lossdict, feature_scores_dict)

    def _save_checkpoint(self,
                         state,
                         is_best,
                         filename="checkpoint"):
        """
        Saves snapshot of the model state to file. Will save a checkpoint
        with name `<filename>.pth.tar` and, if this is the model's best
        performance so far, will save the state to a `best_model.pth.tar`
        file as well.

        Models are saved in the state dictionary format. This is a more
        stable format compared to saving the whole model (which is another
        option supported by PyTorch). Note that we do save a number of
        additional, Selene-specific parameters in the dictionary
        and that the actual `model.state_dict()` is stored in the `state_dict`
        key of the dictionary loaded by `torch.load`.

        See: https://pytorch.org/docs/stable/notes/serialization.html for more
        information about how models are saved in PyTorch.

        Parameters
        ----------
        state : dict
            Information about the state of the model. Note that this is
            not `model.state_dict()`, but rather, a dictionary containing
            keys that can be used for continued training in Selene
            _in addition_ to a key `state_dict` that contains
            `model.state_dict()`.
        is_best : bool
            Is this the model's best performance so far?
        filename : str, optional
            Default is "checkpoint". Specify the checkpoint filename. Will
            append a file extension to the end of the `filename`
            (e.g. `checkpoint.pth.tar`).

        Returns
        -------
        None

        """
        logger.debug("[TRAIN] {0}: Saving model state to file.".format(
            state["step"]))
        cp_filepath = os.path.join(
            self.output_dir, filename)
        torch.save(state, "{0}.pth.tar".format(cp_filepath))
        if is_best:
            best_filepath = os.path.join(self.output_dir, "best_model")
            shutil.copyfile("{0}.pth.tar".format(cp_filepath),
                            "{0}.pth.tar".format(best_filepath))



"""
This module provides the NonStrandSpecific class.
"""
import torch
from torch.nn.modules import Module

from . import _is_lua_trained_model


def _flip(x, dim):
    """
    Reverses the elements in a given dimension `dim` of the Tensor.

    source: https://github.com/pytorch/pytorch/issues/229
    """
    xsize = x.size()
    dim = x.dim() + dim if dim < 0 else dim
    x = x.contiguous()
    x = x.view(-1, *xsize[dim:])
    x = x.view(
        x.size(0), x.size(1), -1)[:, getattr(
            torch.arange(x.size(1)-1, -1, -1),
            ('cpu','cuda')[x.is_cuda])().long(), :]
    return x.view(xsize)


class NonStrandSpecific(Module):
    """
    A torch.nn.Module that wraps a user-specified model architecture if the
    architecture does not need to account for sequence strand-specificity.

    Parameters
    ----------
    model : torch.nn.Module
        The user-specified model architecture.
    mode : {'mean', 'max'}, optional
        Default is 'mean'. NonStrandSpecific will pass the input and the
        reverse-complement of the input into `model`. The mode specifies
        whether we should output the mean or max of the predictions as
        the non-strand specific prediction.

    Attributes
    ----------
    model : torch.nn.Module
        The user-specified model architecture.
    mode : {'mean', 'max'}
        How to handle outputting a non-strand specific prediction.

    """

    def __init__(self, model, mode="mean"):
        super(NonStrandSpecific, self).__init__()

        self.model = model

        if mode != "mean" and mode != "max":
            raise ValueError("Mode should be one of 'mean' or 'max' but was"
                             "{0}.".format(mode))
        self.mode = mode
        self.from_lua = _is_lua_trained_model(model)

    def forward(self, input):
        reverse_input = None
        if self.from_lua:
            reverse_input = _flip(
                _flip(torch.squeeze(input, 2), 1), 2).unsqueeze_(2)
        else:
            reverse_input = _flip(_flip(input, 1), 2)

        output = self.model.forward(input)
        output_from_rev = self.model.forward(reverse_input)
        if type(output) == tuple:
            output1, output2 = output
            output_from_rev1, output_from_rev2 = output_from_rev
            if self.mode == "mean":
                return ((output1 + output_from_rev1) / 2,
                        (output2 + output_from_rev2) / 2)
            else:
                return (torch.max(output1, output_from_rev1),
                        torch.max(output2, output_from_rev2))
        else:
            if self.mode == "mean":
                return (output + output_from_rev) / 2
            else:
                return torch.max(output, output_from_rev)





"""Command line interface for Selene.

Why does this file exist, and why not put this in ``__main__``? You might be tempted to import things from ``__main__``
later, but that will cause problems--the code will get executed twice:

- When you run ``python3 -m selene_sdk`` python will execute``__main__.py`` as a script. That means there won't be any
  ``selene_sdk.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``selene_sdk.__main__`` in ``sys.modules``.

.. seealso:: http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import click

from selene_sdk import __version__
from selene_sdk.utils import load_and_parse_configs_and_run


@click.command()
@click.version_option(__version__)
@click.argument('path', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('--lr', type=float, help='If training, the optimizer learning rate', show_default=True)
def main(path, lr):
    """Build the model and trains it using user-specified input data."""
    load_and_parse_configs_and_run(path, lr=lr)


if __name__ == "__main__":
    main()



"""
This module provides the `MultiSampler` class, which accepts
either an online sampler or a file sampler for each mode of
sampling (train, test, validation).
MultiSampler is a subclass of Sampler.
"""
import numpy as np
from torch.utils.data import DataLoader

from .sampler import Sampler
from .file_samplers import FileSampler


def MultiFileSampler(*args, **kwargs):
    """
    `MultiFileSampler` is deprecated and will be removed from future
    versions of Selene. Please use `MultiSampler` instead. This function
    maintains backward compatibility for code that uses `MultiFileSampler`,
    but we will remove this function in future. Please refer to the
    `MultiSampler` documentation for usage.
    """
    from warnings import warn
    warn("MultiFileSampler is deprecated and will be removed from future "
    "versions of Selene. Please use MultiSampler instead.")
    return MultiSampler(*args, **kwargs)


class MultiSampler(Sampler):
    """
    This sampler draws samples from individual file samplers or data loaders
    that corresponds to training, validation, and testing (optional) modes.
    MultiSampler calls on the correct file sampler or data loader to draw
    samples for a given mode. Example file samplers are under
    `selene_sdk.samplers.file_samplers` and example data loaders are under
    `selene_sdk.samplers.dataloaders`.

    MultiSampler can use either file samplers or data loaders for
    different modes. Using data loaders for some modes while using file samplers
    for other modes are also allowed. The file samplers parse data files
    (e.g. bed, mat, or hdf5). The data loaders provide multi-worker iterators
    that draw samples from online samplers (i.e. on-the-fly sampling). As data
    loaders support parallel sampling, they are generally recommended for
    sampling speed.

    Parameters
    ----------
    train_sampler : selene_sdk.samplers.file_samplers.FileSampler or \
                    selene_sdk.samplers.dataloader.DataLoader
        Load your training data as a `FileSampler` or `DataLoader`
    validate_sampler : FileSampler or DataLoader
        The validation dataset file sampler or data loader.
    features : list(str)
        The list of features the model should predict
    test_sampler : None or FileSampler or DataLoader, optional
        Default is None. The test file sampler is optional.
    mode : str, optional
        Default is "train". Must be one of `{train, validate, test}`. The
        starting mode in which to run the sampler.
    save_datasets : list(str) or None, optional
        Default is None. Currently, we are only including this parameter
        so that `MultiSampler` is consistent with the `Sampler` interface.
        The save dataset functionality for MultiSampler has not been
        defined yet.
    output_dir : str or None, optional
        Default is None. Only used if the sampler has any data or
        logging statements to save to file. Currently not used in
        `MultiSampler`.

    Attributes
    ----------
    modes : list(str)
        A list of the modes that the object may operate in.
    mode : str or None
        Default is `None`. The current mode that the object is operating in.

    """
    def __init__(self,
                 train_sampler,
                 validate_sampler,
                 features,
                 test_sampler=None,
                 mode="train",
                 save_datasets=[],
                 output_dir=None):
        """
        Constructs a new `MultiSampler` object.
        """
        super(MultiSampler, self).__init__(
            features,
            save_datasets=save_datasets,
            output_dir=output_dir)
        self._samplers = {
            "train": train_sampler if (isinstance(train_sampler, FileSampler) or
                                       isinstance(train_sampler, Sampler)) \
                     else None,
            "validate": validate_sampler if (isinstance(validate_sampler, FileSampler) or
                                             isinstance(validate_sampler, Sampler)) \
                        else None
        }

        self._dataloaders = {
            "train": train_sampler if isinstance(train_sampler, DataLoader) \
                     else None,
            "validate": validate_sampler if isinstance(validate_sampler, DataLoader) \
                        else None
        }

        self._iterators = {
            "train": iter(self._dataloaders["train"]) \
                if self._dataloaders["train"] else None,
            "validate": iter(self._dataloaders["validate"]) \
                if self._dataloaders["validate"] else None
        }

        self._index_to_feature = {i: f for (i, f) in enumerate(features)}

        if test_sampler is not None:
            self.modes.append("test")
            self._samplers["test"] = \
                test_sampler if (isinstance(test_sampler, FileSampler) or
                                 isinstance(test_sampler, Sampler)) else None
            self._dataloaders["test"] = \
                test_sampler if isinstance(test_sampler, DataLoader) else None
            self._iterators["test"] = iter(self._dataloaders["test"]) \
                if self._dataloaders["test"] else None

        self.mode = mode

    def set_mode(self, mode):
        """
        Sets the sampling mode.

        Parameters
        ----------
        mode : str
            The name of the mode to use. It must be one of
            `Sampler.BASE_MODES` ("train", "validate") or "test" if
            the test data is supplied.

        Raises
        ------
        ValueError
            If `mode` is not a valid mode.

        """
        if mode not in self.modes:
            raise ValueError(
                "Tried to set mode to be '{0}' but the only valid modes are "
                "{1}".format(mode, self.modes))
        self.mode = mode

    def _set_batch_size(self, batch_size, mode=None):
        """
        Sets the batch size for DataLoader for the specified mode,
        if the specified batch_size does not equal the current batch_size.
        Parameters
        ----------
        batch_size : int
            The batch size for the mode.
        mode : str, optional
            Default is None. The  mode to set batch_size
            If None, will use the current mode `self.mode`.
        """
        if mode is None:
            mode = self.mode

        if self._dataloaders[mode]:
            batch_size_matched = True
            if self._dataloaders[mode].batch_sampler:
                if self._dataloaders[mode].batch_sampler.batch_size != batch_size:
                    self._dataloaders[mode].batch_sampler.batch_size = batch_size
                    batch_size_matched = False
            else:
                if self._dataloaders[mode].batch_size != batch_size:
                    self._dataloaders[mode].batch_size = batch_size
                    batch_size_matched = False

            if not batch_size_matched:
                print("Reset data loader for mode {0} to use the new batch "
                      "size: {1}.".format(mode, batch_size))
                self._iterators[mode] = iter(self._dataloaders[mode])

    def get_feature_from_index(self, index):
        """
        Returns the feature corresponding to an index in the feature
        vector.

        Parameters
        ----------
        index : int
            The index of the feature to retrieve the name for.

        Returns
        -------
        str
            The name of the feature occurring at the specified index.

        """
        return self._index_to_feature[index]

    def sample(self, batch_size=1, mode=None):
        """
        Fetches a mini-batch of the data from the sampler.

        Parameters
        ----------
        batch_size : int, optional
            Default is 1. The size of the batch to retrieve.
        mode : str, optional
            Default is None. The operating mode that the object should run in.
            If None, will use the current mode `self.mode`.
        """
        mode = mode if mode else self.mode
        if self._samplers[mode]:
            return self._samplers[mode].sample(batch_size)
        else:
            self._set_batch_size(batch_size, mode=mode)
            try:
                data, targets = next(self._iterators[mode])
                return data.numpy(), targets.numpy()
            except StopIteration:
                #If DataLoader iterator reaches its length, reinitialize
                self._iterators[mode] = iter(self._dataloaders[mode])

                data, targets = next(self._iterators[mode])
                return data.numpy(), targets.numpy()

    def get_data_and_targets(self, batch_size, n_samples=None, mode=None):
        """
        This method fetches a subset of the data from the sampler,
        divided into batches. This method also allows the user to
        specify what operating mode to run the sampler in when fetching
        the data.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int or None, optional
            Default is None. The total number of samples to retrieve.
            If `n_samples` is None, if a FileSampler is specified for the
            mode, the number of samplers returned is defined by the FileSampler,
            or if a Dataloader is specified, will set `n_samples` to 32000
            if the mode is `validate`, or 640000 if the mode is `test`.
            If the mode is `train` you must have specified a value for
            `n_samples`.
        mode : str, optional
            Default is None. The operating mode that the object should run in.
            If None, will use the current mode `self.mode`.
        """
        mode = mode if mode else self.mode
        if self._samplers[mode]:
            return self._samplers[mode].get_data_and_targets(
                batch_size, n_samples)
        else:
            if n_samples is None:
                if mode == 'validate':
                    n_samples = 32000
                elif mode == 'test':
                    n_samples = 640000
            self._set_batch_size(batch_size, mode=mode)
            data_and_targets = []
            targets_mat = []
            for s in range(0, n_samples, batch_size):
                e = min(n_samples, s+batch_size)
                data, targets = self.sample(batch_size=batch_size, mode=mode)
                data_and_targets.append((data[:e-s], targets[:e-s]))
                targets_mat.append(targets[:e-s])
            targets_mat = np.vstack(targets_mat)
            return data_and_targets, targets_mat

    def get_validation_set(self, batch_size, n_samples=None):
        """
        This method returns a subset of validation data from the
        sampler, divided into batches.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int, optional
            Default is None. The total number of validation examples to
            retrieve. If `n_samples` is None,
            then if a FileSampler is specified for the 'validate' mode, the
            number of samplers returned is defined by the FileSample,
            or if a Dataloader is specified, will set `n_samples` to
            32000.

        Returns
        -------
        sequences_and_targets, targets_matrix : \
        tuple(list(tuple(numpy.ndarray, numpy.ndarray)), numpy.ndarray)
            Tuple containing the list of sequence-target pairs, as well
            as a single matrix with all targets in the same order.
            Note that `sequences_and_targets` sequence elements are of
            the shape :math:`B \\times L \\times N` and its target
            elements are of the shape :math:`B \\times F`, where
            :math:`B` is `batch_size`, :math:`L` is the sequence length,
            :math:`N` is the size of the sequence type's alphabet, and
            :math:`F` is the number of features. Further,
            `target_matrix` is of the shape :math:`S \\times F`, where
            :math:`S = n_samples`.

        Raises
        ------
        ValueError
            If no test partition of the data was specified during
            sampler initialization.
        """
        return self.get_data_and_targets(
            batch_size, n_samples, mode="validate")

    def get_test_set(self, batch_size, n_samples=None):
        """
        This method returns a subset of testing data from the
        sampler, divided into batches.

        Parameters
        ----------
        batch_size : int
            The size of the batches to divide the data into.
        n_samples : int or None, optional
            Default is None. The total number of test examples to
            retrieve. If `n_samples` is None,
            then if a FileSampler is specified for the 'test' mode, the
            number of samplers returned is defined by the FileSample,
            or if a Dataloader is specified, will set `n_samples` to
            640000.

        Returns
        -------
        sequences_and_targets, targets_matrix : \
        tuple(list(tuple(numpy.ndarray, numpy.ndarray)), numpy.ndarray)
            Tuple containing the list of sequence-target pairs, as well
            as a single matrix with all targets in the same order.
            Note that `sequences_and_targets` sequence elements are of
            the shape :math:`B \\times L \\times N` and its target
            elements are of the shape :math:`B \\times F`, where
            :math:`B` is `batch_size`, :math:`L` is the sequence length,
            :math:`N` is the size of the sequence type's alphabet, and
            :math:`F` is the number of features. Further,
            `target_matrix` is of the shape :math:`S \\times F`, where
            :math:`S = n_samples`.

        Raises
        ------
        ValueError
            If no test partition of the data was specified during
            sampler initialization.
        """
        return self.get_data_and_targets(
            batch_size, n_samples, mode="test")

    def save_dataset_to_file(self, mode, close_filehandle=False):
        """
        We implement this function in this class only because the
        TrainModel class calls this method. In the future, we will
        likely remove this method or implement a different way
        of "saving the data" for file samplers. For example, we
        may only output the row numbers sampled so that users may
        reproduce exactly what order the data was sampled.

        Parameters
        ----------
        mode : str
            Must be one of the modes specified in `save_datasets` during
            sampler initialization.
        close_filehandle : bool, optional
            Default is False. `close_filehandle=True` assumes that all
            data corresponding to the input `mode` has been saved to
            file and `save_dataset_to_file` will not be called with
            `mode` again.
        """
        return None
